@@ -21,6 +21,7 @@
 
 #include "ephy-history-service-private.h"
 #include "ephy-history-types.h"
+#include "ephy-history-type-builtins.h"
 #include "ephy-sqlite-connection.h"
 
 typedef gboolean (*EphyHistoryServiceMethod)                              (EphyHistoryService *self, gpointer data, gpointer *result);
@@ -60,6 +61,7 @@ typedef struct _EphyHistoryServiceMessage {
   gboolean success;
   gpointer result;
   gpointer user_data;
+  GCancellable *cancellable;
   GDestroyNotify method_argument_cleanup;
   EphyHistoryJobCallback callback;
 } EphyHistoryServiceMessage;
@@ -124,15 +126,15 @@ ephy_history_service_finalize (GObject *self)
 }
 
 static gboolean
-impl_visit_url (EphyHistoryService *self, const char *url)
+impl_visit_url (EphyHistoryService *self, const char *url, EphyHistoryPageVisitType visit_type)
 {
   EphyHistoryPageVisit *visit;
 
   visit = ephy_history_page_visit_new (url,
                                        time (NULL),
-                                       EPHY_PAGE_VISIT_TYPED);
+                                       visit_type);
   ephy_history_service_add_visit (self,
-                                  visit, NULL, NULL);
+                                  visit, NULL, NULL, NULL);
   ephy_history_page_visit_free (visit);
 
   return FALSE;
@@ -157,8 +159,9 @@ ephy_history_service_class_init (EphyHistoryServiceClass *klass)
                   g_signal_accumulator_true_handled, NULL,
                   g_cclosure_marshal_generic,
                   G_TYPE_BOOLEAN,
-                  1,
-                  G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);
+                  2,
+                  G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE,
+                  EPHY_TYPE_HISTORY_PAGE_VISIT_TYPE);
 
   signals[CLEARED] =
     g_signal_new ("cleared",
@@ -390,6 +393,7 @@ ephy_history_service_message_new (EphyHistoryService *service,
                                   EphyHistoryServiceMessageType type,
                                   gpointer method_argument,
                                   GDestroyNotify method_argument_cleanup,
+                                  GCancellable *cancellable,
                                   EphyHistoryJobCallback callback,
                                   gpointer user_data)
 {
@@ -399,6 +403,7 @@ ephy_history_service_message_new (EphyHistoryService *service,
   message->type = type;
   message->method_argument = method_argument;
   message->method_argument_cleanup = method_argument_cleanup;
+  message->cancellable = cancellable ? g_object_ref (cancellable) : NULL;
   message->callback = callback;
   message->user_data = user_data;
 
@@ -411,6 +416,9 @@ ephy_history_service_message_free (EphyHistoryServiceMessage *message)
   if (message->method_argument_cleanup)
     message->method_argument_cleanup (message->method_argument);
 
+  if (message->cancellable)
+    g_object_unref (message->cancellable);
+
   g_slice_free1 (sizeof (EphyHistoryServiceMessage), message);
 }
 
@@ -420,6 +428,11 @@ ephy_history_service_execute_job_callback (gpointer data)
   EphyHistoryServiceMessage *message = (EphyHistoryServiceMessage*) data;
 
   g_assert (message->callback);
+
+  if (g_cancellable_is_cancelled (message->cancellable)) {
+    ephy_history_service_message_free (message);
+    return FALSE;
+  }
   message->callback (message->service, message->success, message->result, message->user_data);
 
   if (message->type == CLEAR)
@@ -546,7 +559,7 @@ ephy_history_service_execute_query_hosts (EphyHistoryService *self,
 }
 
 void
-ephy_history_service_add_visit (EphyHistoryService *self, EphyHistoryPageVisit *visit, EphyHistoryJobCallback callback, gpointer user_data)
+ephy_history_service_add_visit (EphyHistoryService *self, EphyHistoryPageVisit *visit, GCancellable *cancellable, EphyHistoryJobCallback callback, gpointer user_data)
 {
   EphyHistoryServiceMessage *message;
 
@@ -556,12 +569,12 @@ ephy_history_service_add_visit (EphyHistoryService *self, EphyHistoryPageVisit *
   message = ephy_history_service_message_new (self, ADD_VISIT,
                                               ephy_history_page_visit_copy (visit),
                                               (GDestroyNotify) ephy_history_page_visit_free,
-                                              callback, user_data);
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
 void
-ephy_history_service_add_visits (EphyHistoryService *self, GList *visits, EphyHistoryJobCallback callback, gpointer user_data)
+ephy_history_service_add_visits (EphyHistoryService *self, GList *visits, GCancellable *cancellable, EphyHistoryJobCallback callback, gpointer user_data)
 {
   EphyHistoryServiceMessage *message;
 
@@ -571,12 +584,12 @@ ephy_history_service_add_visits (EphyHistoryService *self, GList *visits, EphyHi
   message = ephy_history_service_message_new (self, ADD_VISITS,
                                               ephy_history_page_visit_list_copy (visits),
                                               (GDestroyNotify) ephy_history_page_visit_list_free,
-                                              callback, user_data);
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
 void
-ephy_history_service_find_visits_in_time (EphyHistoryService *self, gint64 from, gint64 to, EphyHistoryJobCallback callback, gpointer user_data)
+ephy_history_service_find_visits_in_time (EphyHistoryService *self, gint64 from, gint64 to, GCancellable *cancellable, EphyHistoryJobCallback callback, gpointer user_data)
 {
   EphyHistoryQuery *query;
 
@@ -586,12 +599,12 @@ ephy_history_service_find_visits_in_time (EphyHistoryService *self, gint64 from,
   query->from = from;
   query->to = to;
 
-  ephy_history_service_query_visits (self, query, callback, user_data);
+  ephy_history_service_query_visits (self, query, cancellable, callback, user_data);
   ephy_history_query_free (query);
 }
 
 void
-ephy_history_service_query_visits (EphyHistoryService *self, EphyHistoryQuery *query, EphyHistoryJobCallback callback, gpointer user_data)
+ephy_history_service_query_visits (EphyHistoryService *self, EphyHistoryQuery *query, GCancellable *cancellable, EphyHistoryJobCallback callback, gpointer user_data)
 {
   EphyHistoryServiceMessage *message;
 
@@ -599,7 +612,7 @@ ephy_history_service_query_visits (EphyHistoryService *self, EphyHistoryQuery *q
   g_return_if_fail (query != NULL);
 
   message = ephy_history_service_message_new (self, QUERY_VISITS,
-                                              ephy_history_query_copy (query), (GDestroyNotify) ephy_history_query_free, callback, user_data);
+                                              ephy_history_query_copy (query), (GDestroyNotify) ephy_history_query_free, cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
@@ -614,7 +627,7 @@ ephy_history_service_execute_query_urls (EphyHistoryService *self, EphyHistoryQu
 }
 
 void
-ephy_history_service_query_urls (EphyHistoryService *self, EphyHistoryQuery *query, EphyHistoryJobCallback callback, gpointer user_data)
+ephy_history_service_query_urls (EphyHistoryService *self, EphyHistoryQuery *query, GCancellable *cancellable, EphyHistoryJobCallback callback, gpointer user_data)
 {
   EphyHistoryServiceMessage *message;
 
@@ -622,12 +635,14 @@ ephy_history_service_query_urls (EphyHistoryService *self, EphyHistoryQuery *que
   g_return_if_fail (query != NULL);
 
   message = ephy_history_service_message_new (self, QUERY_URLS,
-                                              ephy_history_query_copy (query), (GDestroyNotify) ephy_history_query_free, callback, user_data);
+                                              ephy_history_query_copy (query), (GDestroyNotify) ephy_history_query_free,
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
 void
 ephy_history_service_get_hosts (EphyHistoryService *self,
+                                GCancellable *cancellable,
                                 EphyHistoryJobCallback callback,
                                 gpointer user_data)
 {
@@ -636,7 +651,7 @@ ephy_history_service_get_hosts (EphyHistoryService *self,
   g_return_if_fail (EPHY_IS_HISTORY_SERVICE (self));
 
   message = ephy_history_service_message_new (self, GET_HOSTS,
-                                              NULL, NULL,
+                                              NULL, NULL, cancellable,
                                               callback, user_data);
   ephy_history_service_send_message (self, message);
 }
@@ -644,6 +659,7 @@ ephy_history_service_get_hosts (EphyHistoryService *self,
 void
 ephy_history_service_query_hosts (EphyHistoryService *self,
                                   EphyHistoryQuery *query,
+                                  GCancellable *cancellable,
                                   EphyHistoryJobCallback callback,
                                   gpointer user_data)
 {
@@ -654,7 +670,7 @@ ephy_history_service_query_hosts (EphyHistoryService *self,
   message = ephy_history_service_message_new (self, QUERY_HOSTS,
                                               ephy_history_query_copy (query),
                                               (GDestroyNotify) ephy_history_query_free,
-                                              callback, user_data);
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
@@ -682,6 +698,7 @@ void
 ephy_history_service_set_url_title (EphyHistoryService *self,
                                     const char *orig_url,
                                     const char *title,
+                                    GCancellable *cancellable,
                                     EphyHistoryJobCallback callback,
                                     gpointer user_data)
 {
@@ -694,7 +711,7 @@ ephy_history_service_set_url_title (EphyHistoryService *self,
   url = ephy_history_url_new (orig_url, title, 0, 0, 0);
   message = ephy_history_service_message_new (self, SET_URL_TITLE,
                                               url, (GDestroyNotify) ephy_history_url_free,
-                                              callback, user_data);
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
@@ -725,6 +742,7 @@ void
 ephy_history_service_set_url_zoom_level (EphyHistoryService *self,
                                          const char *url,
                                          const double zoom_level,
+                                         GCancellable *cancellable,
                                          EphyHistoryJobCallback callback,
                                          gpointer user_data)
 {
@@ -738,7 +756,7 @@ ephy_history_service_set_url_zoom_level (EphyHistoryService *self,
 
   message = ephy_history_service_message_new (self, SET_URL_ZOOM_LEVEL,
                                               variant, (GDestroyNotify)g_variant_unref,
-                                              callback, user_data);
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
@@ -759,6 +777,7 @@ ephy_history_service_execute_get_url (EphyHistoryService *self,
 void
 ephy_history_service_get_url (EphyHistoryService *self,
                               const char *url,
+                              GCancellable *cancellable,
                               EphyHistoryJobCallback callback,
                               gpointer user_data)
 {
@@ -769,7 +788,7 @@ ephy_history_service_get_url (EphyHistoryService *self,
 
   message =  ephy_history_service_message_new (self, GET_URL,
                                                g_strdup (url), g_free,
-                                               callback, user_data);
+                                               cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
@@ -791,6 +810,7 @@ ephy_history_service_execute_get_host_for_url (EphyHistoryService *self,
 void
 ephy_history_service_get_host_for_url (EphyHistoryService *self,
                                        const char *url,
+                                       GCancellable *cancellable,
                                        EphyHistoryJobCallback callback,
                                        gpointer user_data)
 {
@@ -801,7 +821,7 @@ ephy_history_service_get_host_for_url (EphyHistoryService *self,
 
   message = ephy_history_service_message_new (self, GET_HOST_FOR_URL,
                                               g_strdup (url), g_free,
-                                              callback, user_data);
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
@@ -851,6 +871,7 @@ ephy_history_service_execute_clear (EphyHistoryService *self,
 void
 ephy_history_service_delete_urls (EphyHistoryService *self,
                                   GList *urls,
+                                  GCancellable *cancellable,
                                   EphyHistoryJobCallback callback,
                                   gpointer user_data)
 {
@@ -861,25 +882,27 @@ ephy_history_service_delete_urls (EphyHistoryService *self,
 
   message = ephy_history_service_message_new (self, DELETE_URLS, 
                                               ephy_history_url_list_copy (urls), (GDestroyNotify)ephy_history_url_list_free,
-                                              callback, user_data);
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
 void
 ephy_history_service_delete_host (EphyHistoryService *self,
                                   EphyHistoryHost *host,
+                                  GCancellable *cancellable,
                                   EphyHistoryJobCallback callback,
                                   gpointer user_data)
 {
   EphyHistoryServiceMessage *message =
     ephy_history_service_message_new (self, DELETE_HOST,
                                       ephy_history_host_copy (host), (GDestroyNotify)ephy_history_host_free,
-                                      callback, user_data);
+                                      cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
 void
 ephy_history_service_clear (EphyHistoryService *self,
+                            GCancellable *cancellable,
                             EphyHistoryJobCallback callback,
                             gpointer user_data)
 {
@@ -889,7 +912,7 @@ ephy_history_service_clear (EphyHistoryService *self,
 
   message = ephy_history_service_message_new (self, CLEAR,
                                               NULL, NULL,
-                                              callback, user_data);
+                                              cancellable, callback, user_data);
   ephy_history_service_send_message (self, message);
 }
 
@@ -900,7 +923,7 @@ ephy_history_service_quit (EphyHistoryService *self,
 {
   EphyHistoryServiceMessage *message =
     ephy_history_service_message_new (self, QUIT, 
-                                      NULL, NULL,
+                                      NULL, NULL, NULL,
                                       callback, user_data);
   ephy_history_service_send_message (self, message);
 }
@@ -922,6 +945,12 @@ static EphyHistoryServiceMethod methods[] = {
   (EphyHistoryServiceMethod)ephy_history_service_execute_query_hosts
 };
 
+static gboolean
+ephy_history_service_message_is_write (EphyHistoryServiceMessage *message)
+{
+  return message->type < QUIT;
+}
+
 static void
 ephy_history_service_process_message (EphyHistoryService *self,
                                       EphyHistoryServiceMessage *message)
@@ -930,6 +959,12 @@ ephy_history_service_process_message (EphyHistoryService *self,
 
   g_assert (self->priv->history_thread == g_thread_self ());
 
+  if (g_cancellable_is_cancelled (message->cancellable) &&
+      !ephy_history_service_message_is_write (message)) {
+    ephy_history_service_message_free (message);
+    return;
+  }
+
   method = methods[message->type];
   message->result = NULL;
   message->success = method (message->service, message->method_argument, &message->result);
@@ -949,6 +984,7 @@ ephy_history_service_find_urls (EphyHistoryService *self,
                                 gint64 from, gint64 to,
                                 guint limit, gint host,
                                 GList *substring_list,
+                                GCancellable *cancellable,
                                 EphyHistoryJobCallback callback,
                                 gpointer user_data)
 {
@@ -967,24 +1003,28 @@ ephy_history_service_find_urls (EphyHistoryService *self,
     query->limit = limit;
 
   ephy_history_service_query_urls (self,
-                                   query, callback, user_data);
+                                   query, cancellable,
+                                   callback, user_data);
+  ephy_history_query_free (query);
 }
 
 void
 ephy_history_service_visit_url (EphyHistoryService *self,
-                                const char *url)
+                                const char *url,
+                                EphyHistoryPageVisitType visit_type)
 {
   gboolean result;
 
   g_return_if_fail (EPHY_IS_HISTORY_SERVICE (self));
   g_return_if_fail (url != NULL);
 
-  g_signal_emit (self, signals[VISIT_URL], 0, url, &result);
+  g_signal_emit (self, signals[VISIT_URL], 0, url, visit_type, &result);
 }
 
 void
 ephy_history_service_find_hosts (EphyHistoryService *self,
                                  gint64 from, gint64 to,
+                                 GCancellable *cancellable,
                                  EphyHistoryJobCallback callback,
                                  gpointer user_data)
 {
@@ -997,6 +1037,7 @@ ephy_history_service_find_hosts (EphyHistoryService *self,
   query->from = from;
   query->to = to;
 
-  ephy_history_service_query_hosts (self,
-                                    query, callback, user_data);
+  ephy_history_service_query_hosts (self, query,
+                                    cancellable, callback, user_data);
+  ephy_history_query_free (query);
 }
